import efidgy
from efidgy import impl

import urllib.parse


__all__ = [
    'geocode',
]


async def geocode(address, env=None):
    if env is None:
        env = efidgy.env
    env = env.override(code='efidgy')
    c = impl.client.AsyncClient(env)
    data = await c.get(
        '/tools/geocode/?address={}'.format(urllib.parse.quote(address)),
    )

    lat = data.get('lat')
    lon = data.get('lon')
    if lat is None or lon is None:
        raise efidgy.exceptions.GeocodeError(address)

    return lat, lon



import importlib


def import_string(dotted_path):
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        raise ImportError(
            '{} doesn\'t look like a module path'.format(dotted_path),
        )

    module = importlib.import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:
        raise ImportError(
            'Module "{}" does not define a "{}" attribute/class'.format(
                module_path,
                class_name,
            ),
        )



import os

from . import models
from . import asyncapi
from . import exceptions
from . import tools


__all__ = [
    asyncapi,
    exceptions,
    'Env',
    models,
    tools,
    '__version__',
]


##
# Do not touch this line.
# See gitlab ci.
#
__version__ = '0.14.2'


class Env:
    def __init__(self, token=None, code=None, unit_system=None):
        self.host = os.environ.get('EFIDGY_HOST', 'console.efidgy.com')
        self.insecure = os.environ.get('EFIDGY_INSECURE', '0') != '0'
        self.token = token
        self.code = code
        self.unit_system = unit_system

    def override(self, **kwargs):
        return Env(**{
            'token': self.token,
            'code': self.code,
            'unit_system': self.unit_system,
            **kwargs,
        })


env = Env(
    token=os.environ.get('EFIDGY_ACCESS_TOKEN', None),
    code=os.environ.get('EFIDGY_CUSTOMER_CODE', None),
)



import efidgy

from . import client


class SyncAllMixin:
    @classmethod
    def all(cls, **kwargs):
        c = client.SyncClient(cls._get_env())
        path = cls._get_path(kwargs)
        ret = []
        for data in c.get(path):
            ret.append(cls.Model._decode(data, **kwargs))
        return ret

    @classmethod
    def filter(cls, **kwargs):
        ret = []
        for o in cls.all(**kwargs):
            matched = True
            for field in o._meta.fields:
                if (
                    field.name in kwargs
                    and getattr(o, field.name) != kwargs[field.name]
                ):
                    matched = False
                    break
            if matched:
                ret.append(o)
        return ret

    @classmethod
    def first(cls, **kwargs):
        ret = cls.filter(**kwargs)
        if not ret:
            return None
        return ret[0]


class SyncGetMixin:
    @classmethod
    def get(cls, **kwargs):
        c = client.SyncClient(cls._get_env())
        path = cls._get_path(kwargs)
        pk_name = cls.Model._meta.primary_key.name
        pk = kwargs.get(pk_name, None)
        assert pk is not None, (
            'Primary key not provided: {}'.format(pk_name),
        )
        data = c.get('{}/{}'.format(path, pk))
        return cls.Model._decode(data, **kwargs)

    def refresh(self):
        pk_name = self._meta.primary_key.name
        pk = getattr(self, pk_name, None)
        kwargs = {
            **self._get_context(),
            pk_name: pk,
        }
        obj = self.get(**kwargs)
        for field in self._meta.fields:
            setattr(self, field.name, getattr(obj, field.name))


class SyncCreateMixin:
    @classmethod
    def create(cls, **kwargs):
        c = client.SyncClient(cls._get_env())
        path = cls._get_path(kwargs)
        obj = cls.Model(**kwargs)
        data = c.post(path, obj._encode())
        return cls.Model._decode(data, **kwargs)


class SyncSaveMixin:
    def save(self):
        c = client.SyncClient(self._get_env())
        path = self._get_path(self._get_context())
        c.put('{}/{}'.format(path, self.pk), self._encode(self))


class SyncDeleteMixin:
    @classmethod
    def _bind(cls, Model):
        super()._bind(Model)
        Model.delete = lambda self: self.service.delete(self)

    def delete(self, obj):
        c = client.SyncClient(self._get_env())
        path = self._get_path(obj._get_context())
        c.delete('{}/{}'.format(path, obj.pk))


class SyncViewMixin(
            SyncAllMixin,
            SyncGetMixin,
        ):
    pass


class SyncChangeMixin(
            SyncCreateMixin,
            SyncSaveMixin,
            SyncDeleteMixin,
            SyncViewMixin
        ):
    pass


class AsyncAllMixin:
    @classmethod
    async def all(cls, **kwargs):
        c = client.AsyncClient(cls._get_env())
        path = cls._get_path(kwargs)
        ret = []
        for data in await c.get(path):
            ret.append(cls.Model._decode(data, **kwargs))
        return ret

    @classmethod
    async def filter(cls, **kwargs):
        ret = []
        for o in await cls.all(**kwargs):
            matched = True
            for field, value in kwargs.items():
                if getattr(o, field, None) != value:
                    matched = False
                    break
            if matched:
                ret.append(o)
        return ret

    @classmethod
    async def first(cls, **kwargs):
        ret = await cls.filter(**kwargs)
        if not ret:
            return None
        return ret[0]


class AsyncGetMixin:
    @classmethod
    async def get(cls, **kwargs):
        c = client.AsyncClient(cls._get_env())
        path = cls._get_path(kwargs)
        pk_name = cls._meta.primary_key.name
        pk = kwargs.get(pk_name, None)
        assert pk is not None, (
            'Primary key not provided: {}'.format(pk_name),
        )
        data = await c.get('{}/{}'.format(path, pk))
        return cls._decode(data, **kwargs)

    async def refresh(self):
        pk_name = self._meta.primary_key.name
        pk = getattr(self, pk_name, None)
        kwargs = {
            **self._get_context(),
            pk_name: pk,
        }
        obj = await self.get(**kwargs)
        for field in self._meta.fields:
            setattr(self, field.name, getattr(obj, field.name))


class AsyncCreateMixin:
    @classmethod
    async def create(cls, **kwargs):
        c = client.AsyncClient(cls._get_env())
        path = cls._get_path(kwargs)
        obj = cls.Model(**kwargs)
        data = await c.post(path, obj._encode())
        return cls.Model._decode(data, **kwargs)


class AsyncSaveMixin:
    async def save(self, **kwargs):
        c = client.AsyncClient(self._get_env())
        path = self._get_path(self._get_context())
        await c.put(
            '{}/{}'.format(path, self.pk),
            self._encode(self),
        )


class AsyncDeleteMixin:
    @classmethod
    def _bind(cls, Model):
        super()._bind(Model)
        Model.delete = lambda self: self.service.delete(self)

    async def delete(self, obj):
        c = client.AsyncClient(self._get_env())
        path = self._get_path(obj._get_context())
        await c.delete('{}/{}'.format(path, obj.pk))


class AsyncViewMixin(
            AsyncAllMixin,
            AsyncGetMixin,
        ):
    pass


class AsyncChangeMixin(
            AsyncCreateMixin,
            AsyncSaveMixin,
            AsyncDeleteMixin,
            AsyncViewMixin,
        ):
    pass


class Service:
    Model = None
    path = None
    _env = None

    def __init__(self):
        assert self.path is not None, (
            'Service path must be defined: {}'.format(self)
        )

    @classmethod
    def _get_path(cls, context):
        return cls.path

    @classmethod
    def _get_env(cls):
        if cls._env is None:
            return efidgy.env
        return cls._env

    @classmethod
    def use(cls, env):
        class ProxyService(cls):
            _env = env
        return ProxyService

    @classmethod
    def _bind(cls, Model):
        cls.Model = Model


class EfidgyService(Service):
    @classmethod
    def _get_env(cls):
        return super()._get_env().override(code='efidgy')


class CustomerService(Service):
    pass


class ProjectService(CustomerService):
    @classmethod
    def _get_path(cls, context):
        project = context.get('project')
        assert project is not None, (
            'Project not passed.'
        )
        return '/projects/{project}{path}'.format(
            project=project.pk,
            path=cls.path,
        )

    def _get_context(self):
        return {
            **super()._get_context(),
            'project': self.project,
        }

    # def __init__(self, project=None, **kwargs):
    #     super().__init__(**kwargs)
    #     assert project is not None, (
    #         'Project not specified.'
    #     )
    #     self.project = project


class SolutionService(ProjectService):
    @classmethod
    def _get_path(cls, context):
        solution = context.get('solution')
        if solution is None:
            return super()._get_path(context)
        project = context['project']
        return '/projects/{project}/solutions/{solution}{path}'.format(
            project=project.pk,
            solution=solution.pk,
            path=cls.path,
        )

    def _get_context(self):
        return {
            **super()._get_context(),
            'solution': self.solution,
        }

    # def __init__(self, solution=None, **kwargs):
    #     super().__init__(**kwargs)
    #     self.solution = solution



from asyncio import sleep

from efidgy import impl
from efidgy import models
from efidgy.models import ProjectState
from efidgy.models import JobState
from efidgy.models import ProjectTypeCode
from efidgy.models import SharedMode
from efidgy.models import UnitSystem

from . import idd_or


__all__ = [
    idd_or,
    # 'Computation',
    # 'Issue',
    # 'IssueStats',
    # 'IssueType',
    # 'Issues',
    # 'JobMessage',
    JobState,
    'Member',
    'Project',
    ProjectState,
    # 'ProjectSummary',
    'ProjectType',
    ProjectTypeCode,
    # 'Role',
    SharedMode,
    'Solution',
    # 'SolutionSummary',
    UnitSystem,
]


class Member(models.IMember):
    pass


class Currency(models.ICurrency):
    class service(impl.service.AsyncViewMixin, impl.service.EfidgyService):
        path = '/refs/currencies'


class ProjectType(models.IProjectType):
    class service(impl.service.AsyncViewMixin, impl.service.EfidgyService):
        path = '/refs/project_types'


class Project(models.IProject):
    currency = impl.fields.ObjectField(model=Currency)
    project_type = impl.fields.ObjectField(model=ProjectType)
    member = impl.fields.ObjectField(model=Member)

    class service(impl.service.AsyncChangeMixin, impl.service.CustomerService):
        path = '/projects'

    async def start_computation(self):
        c = impl.client.AsyncClient(self.service._get_env())
        await c.post(self._get_computation_path(), None)

    async def stop_computation(self):
        c = impl.client.AsyncClient(self.service._get_env())
        await c.delete(self._get_computation_path())

    async def wait_computation(self):
        while True:
            c = impl.client.AsyncClient(self.service._get_env())
            response = await c.get(self._get_computation_path())
            self._log_messages(response['messages'])
            if response['state'] not in [JobState.PENDING, JobState.WORKING]:
                break
            await sleep(10)

    async def computate(self):
        await self.start_computation()
        await self.wait_computation()


class Solution(models.ISolution):
    class service(impl.service.AsyncViewMixin, impl.service.ProjectService):
        path = '/solutions'



import os
import re
from setuptools import setup
from setuptools import find_packages


def read(fname):
    return open(fname, encoding='utf8').read()


def version():
    for line in open('efidgy/__init__.py'):
        m = re.match(r'__version__ = \'(.*)\'', line)
        if m:
            return m[1]
    raise AssertionError('No version found.')


setup(
    name='efidgy',
    version=version(),
    author='Vasily Stepanov',
    author_email='vasily.stepanov@efidgy.com',
    license='GPL3',
    keywords='efidgy logistics optimization dispatcing vrp',
    url='https://efidgy.com/',
    description='Python bindings to efidgy services.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries',
    ],
    project_urls={
        'Documentation': 'https://efidgy.com/docs',
        'Source': 'https://github.com/efidgy/efidgy',
    },
    packages=find_packages(),
    install_requires=[
    ],
    python_requires='>=3.6',
)



from efidgy import impl
from efidgy import models
from efidgy.models.idd_or import PointType


__all__ = [
    'Order',
    'OrderRoute',
    'OrderSchedule',
    'Path',
    'Point',
    PointType,
    'Store',
    'Vehicle',
    'VehicleRoute',
    'VehicleSchedule',
]


class Point(models.idd_or.IPoint):
    pass


class Path(models.idd_or.IPath):
    pass


class Store(models.idd_or.IStore):
    class service(impl.service.AsyncChangeMixin, impl.service.ProjectService):
        path = '/stores'


class Vehicle(models.idd_or.IVehicle):
    store = impl.fields.ObjectField(model=Store)
    route = impl.fields.ObjectField(model='efidgy.models.idd_or.VehicleRoute')

    class service(impl.service.AsyncChangeMixin, impl.service.SolutionService):
        path = '/vehicles'


class VehicleRoute(models.idd_or.IVehicleRoute):
    pass


class VehicleSchedule(models.idd_or.IVehicleSchedule):
    start_point = impl.fields.PolymorphObjectField(
        lookup_field='point_type',
        models={
            PointType.STORE: 'efidgy.models.idd_or.Store',
            PointType.ORDER: 'efidgy.models.idd_or.Order',
        },
    )
    end_point = impl.fields.PolymorphObjectField(
        lookup_field='point_type',
        models={
            PointType.STORE: 'efidgy.models.idd_or.Store',
            PointType.ORDER: 'efidgy.models.idd_or.Order',
        },
    )
    path = impl.fields.ObjectField(model=Path)
    orders = impl.fields.ListField(item='efidgy.models.idd_or.Order')


class Order(models.idd_or.IOrder):
    store = impl.fields.ObjectField(model=Store)

    class service(impl.service.AsyncChangeMixin, impl.service.SolutionService):
        path = '/orders'


class OrderRoute(models.idd_or.IOrderRoute):
    vehicle = impl.fields.ObjectField(
        model='efidgy.asyncapi.models.idd_or.Vehicle',
    )
    schedule = impl.fields.ListField(item='efidgy.models.idd_or.OrderSchedule')


class OrderSchedule(models.idd_or.IOrderSchedule):
    start_point = impl.fields.PolymorphObjectField(
        lookup_field='point_type',
        models={
            PointType.STORE: Store,
            PointType.ORDER: Order,
        },
    )
    end_point = impl.fields.PolymorphObjectField(
        lookup_field='point_type',
        models={
            PointType.STORE: Store,
            PointType.ORDER: Order,
        },
    )
    path = impl.fields.ObjectField(model=Path)



import datetime

import re

from math import floor

from . import routines


class Field:
    def __init__(self, primary_key=False, required=True):
        self.required = required
        self.primary_key = primary_key

    def _decode(self, value, **kwargs):
        return value

    def _encode(self, value):
        return value


class BooleanField(Field):
    pass


class FloatField(Field):
    pass


class IntegerField(Field):
    pass


class CharField(Field):
    pass


class PrimaryKey(CharField):
    def __init__(self, **kwargs):
        kwargs['primary_key'] = True
        super().__init__(**kwargs)


class TimeField(Field):
    def _decode(self, value, **kwargs):
        if value is None:
            return None
        m = re.match(r'(\d{2}):(\d{2})', value)
        assert m, (
            'Unexpected time value: {}'.format(value)
        )
        return datetime.time(int(m[1]), int(m[2]))

    def _encode(self, value):
        if value is None:
            return None
        return '{:02d}:{:02d}'.format(value.hour, value.minute)


class DurationField(Field):
    def _decode(self, value, **kwargs):
        if value is None:
            return None
        m = re.match(r'(\d+):(\d{2})', value)
        assert m, (
            'Unexpected duration value: {}'.format(value)
        )
        return datetime.timedelta(hours=int(m[1]), minutes=int(m[2]))

    def _encode(self, value):
        if value is None:
            return None
        seconds = value.total_seconds()
        hours = int(floor(seconds / 3600))
        minutes = int(floor(seconds / 60) % 60)
        return '{:d}:{:02d}'.format(hours, minutes)


class DictField(Field):
    pass


class ObjectField(Field):
    def __init__(self, model=None, **kwargs):
        super().__init__(**kwargs)
        self._model = model

    @property
    def model(self):
        if isinstance(self._model, str):
            self._model = routines.import_string(self._model)
        return self._model

    def _decode(self, value, **kwargs):
        if value is None:
            return None
        return self.model._decode(value, **kwargs)

    def _encode(self, value):
        if value is None:
            return None

        if not isinstance(value, self.model):
            primary_key = self.model._meta.primary_key
            if primary_key is None:
                raise AssertionError(
                    '{} instance expected: {}'.format(self.model, type(value))
                )
            value = self.model(**{primary_key.name: value})

        return self.model._encode(value)


class PolymorphObjectField(Field):
    def __init__(self, lookup_field=None, models=None, **kwargs):
        super().__init__(**kwargs)
        self.lookup_field = lookup_field
        self.models = models

    def _get_model(self, value):
        object_type = value.get(self.lookup_field)
        assert object_type is not None, (
            'Lookup field not found: {}'.format(self.lookup_field)
        )

        model = self.models.get(object_type)
        assert model is not None, (
            'Model not found: {}'.format(object_type)
        )

        if isinstance(model, str):
            model = routines.import_string(model)
            self.models[object_type] = model

        return model

    def _decode(self, value, **kwargs):
        if value is None:
            return None

        return self._get_model(value)._decode(value, **kwargs)

    def _encode(self, value):
        if value is None:
            return None

        model = self._get_model(value)

        if not isinstance(value, model):
            primary_key = self.model._meta.primary_key
            if primary_key is None:
                raise AssertionError(
                    '{} instance expected: {}'.format(model, type(value))
                )
            value = self.model(**{primary_key.name: value})

        return model._encode(value)


class ListField(Field):
    def __init__(self, item=None, **kwargs):
        super().__init__(**kwargs)
        self._item = item

    @property
    def item(self):
        if isinstance(self._item, str):
            self._item = routines.import_string(self._item)
        return self._item

    def _decode(self, value, **kwargs):
        if value is None:
            return None
        ret = []
        for item in value:
            ret.append(self.item._decode(item, **kwargs))
        return ret

    def _encode(self, value):
        if value is None:
            return None
        ret = []
        for item in value:
            ret.append(self.item._encode(item))
        return ret
